"""
Hyperparameter descriptor for reconstruction methods.
"""

from __future__ import annotations

import numbers
from typing import Any


class Hyperparameter:
    """
    Descriptor class for hyperparameters with validation logic.

    This descriptor validates hyperparameters by type,
    bounds (for numeric types), or valid values (for categorical types)
    when they are accessed or assigned.

    Parameters
    ----------
    param_type : type
        The expected type for the hyperparameter (int, float, str, etc.).
    bounds : tuple, optional
        For numeric types, a tuple of (min_value, max_value) bounds.
    values : list, optional
        For categorical types, a list of valid values.
    default : Any, optional
        Default value for the hyperparameter.

    Examples
    --------
    >>> class SomeReconstructor:
    ...     power = Hyperparameter(float, bounds=(0.5, 5.0), default=2.0)
    ...     k = Hyperparameter(int, bounds=(1, 20), default=5)
    ...     mode = Hyperparameter(str, values=['fast', 'slow'], default='fast')
    """

    def __init__(
        self,
        param_type: type,
        *,
        bounds: tuple[int | float, int | float] | None = None,
        values: list[int | float | str] | None = None,
        default: Any = None,
    ):
        self.param_type = param_type
        self.bounds = bounds
        self.values = values
        self.default = default
        self.name = None
        self.private_name = None

        if bounds is not None and values is not None:
            raise ValueError("Cannot specify both bounds and values")

        if bounds is not None:
            if not (
                isinstance(param_type, type)
                and issubclass(param_type, numbers.Number)
            ):
                raise ValueError(
                    "Bounds can only be specified for numeric types"
                )
            if len(bounds) != 2:
                raise ValueError(
                    "Bounds must be a tuple of (min_value, max_value)"
                )
            if bounds[0] is not None and bounds[1] is not None:
                if bounds[0] >= bounds[1]:
                    raise ValueError(
                        "Lower bound must be less than upper bound"
                    )

    def __set_name__(self, owner, name):
        """Called when the descriptor is assigned to a class attribute."""
        self.name = name
        self.private_name = f"_{name}"

    def __get__(self, instance, owner):
        """Get the hyperparameter value."""
        if instance is None:
            return self

        return getattr(instance, self.private_name, self.default)

    def __set__(self, instance, value):
        """Set the hyperparameter value with validation."""
        validated_value = self._validate_and_cast(value)
        setattr(instance, self.private_name, validated_value)

    def _validate_and_cast(self, value):
        """Validate and cast the value according to the hyperparameter specification."""
        if value is None:
            return self.default

        try:
            casted_value = self.param_type(value)
        except (ValueError, TypeError) as e:
            raise TypeError(
                f"Cannot convert {value!r} to {self.param_type.__name__} for parameter '{self.name}'"
            ) from e

        if self.bounds is not None and isinstance(
            casted_value, numbers.Number
        ):
            min_val, max_val = self.bounds
            if min_val is not None and casted_value < min_val:
                raise ValueError(
                    f"Parameter '{self.name}' value {casted_value} is below the minimum bound {min_val}"
                )
            if max_val is not None and casted_value > max_val:
                raise ValueError(
                    f"Parameter '{self.name}' value {casted_value} is above the maximum bound {max_val}"
                )

        if self.values is not None:
            if casted_value not in self.values:
                raise ValueError(
                    f"Parameter '{self.name}' value {casted_value!r} not in valid values {self.values}"
                )

        return casted_value

    def get_spec(self) -> dict[str, Any]:
        """
        Get the hyperparameter specification as a dictionary.

        Returns
        -------
        dict[str, Any]
            Dictionary containing the hyperparameter specification with keys:
            - 'type': The parameter type
            - 'bounds': Bounds tuple (if specified)
            - 'values': List of valid values (if specified)
            - 'default': Default value (if specified)
        """
        spec = {"type": self.param_type}

        if self.bounds is not None:
            spec["bounds"] = self.bounds

        if self.values is not None:
            spec["values"] = self.values

        if self.default is not None:
            spec["default"] = self.default

        return spec



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import ClassVar

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from climatrix import BaseClimatrixDataset, Domain
from climatrix.decorators.runtime import log_input, raise_if_not_installed
from climatrix.reconstruct.base import BaseReconstructor
from climatrix.reconstruct.sinet.callbacks import EarlyStopping
from climatrix.reconstruct.sinet.dataset import (
    SiNETDatasetGenerator,
)
from climatrix.reconstruct.sinet.losses import LossEntity, compute_sdf_losses
from climatrix.reconstruct.sinet.model import SiNET
from climatrix.utils.hyperparameter import Hyperparameter

log = logging.getLogger(__name__)


class SiNETReconstructor(BaseReconstructor):
    """
    Spatial Interpolation Network (SiNET) Reconstructor.

    SiNET is a neural network-based method for spatial interpolation that
    uses implicit neural representations to reconstruct continuous fields
    from sparse observations.

    Parameters
    ----------
    dataset : BaseClimatrixDataset
        Source dataset to reconstruct from.
    target_domain : Domain
        Target domain to reconstruct onto.
    layers : int, optional
        Number of hidden layers in the network (default is 2).
    hidden_dim : int, optional
        Number of neurons in each hidden layer (default is 64).
    sorting_group_size : int, optional
        Size of sorting groups for data processing (default is 16).
    scale : float, optional
        Scaling factor for coordinates (default is 1.5).
    lr : float, optional
        Learning rate for optimization (default is 3e-4).
        Type: float, bounds: <unbounded>, default: 1e-3
    batch_size : int, optional
        Batch size for training (default is 512).
        Type: int, bounds: <unbounded>, default: 128
    num_epochs : int, optional
        Number of training epochs (default is 5000).
        Type: int, bounds: <unbounded>, default: 5_000
    num_workers : int, optional
        Number of worker processes for data loading (default is 0).
    device : str, optional
        Device to run computation on (default is "cuda").
    gradient_clipping_value : float | None, optional
        Value for gradient clipping (default is None).
        Type: float, bounds: <unbounded>, default: 1.0
    checkpoint : str | os.PathLike | Path | None, optional
        Path to model checkpoint (default is None).
    mse_loss_weight : float, optional
        Weight for MSE loss component (default is 3e3).
        Type: float, bounds: <unbounded>, default: 1e2
    eikonal_loss_weight : float, optional
        Weight for Eikonal loss component (default is 5e1).
        Type: float, bounds: <unbounded>, default: 1e1
    laplace_loss_weight : float, optional
        Weight for Laplace loss component (default is 1e2).
        Type: float, bounds: <unbounded>, default: 1e2
    validation : float | BaseClimatrixDataset, optional
        Validation data or portion for training (default is 0.2).
    patience : int | None, optional
        Early stopping patience (default is None).
    overwrite_checkpoint : bool, optional
        Whether to overwrite existing checkpoints (default is False).

    Raises
    ------
    ValueError
        If SiNET is used with dynamic datasets or if CUDA is not available
        when requested.

    Notes
    -----
    Hyperparameters for optimization:
        - lr: float in (1e-5, 1e-2), default=1e-3
        - batch_size: int in (64, 1024), default=128
        - num_epochs: int in (10, 10_000), default=5_000
        - gradient_clipping_value: float in (0.1, 10.0), default=1.0
        - mse_loss_weight: float in (1e1, 1e4), default=1e2
        - eikonal_loss_weight: float in (1e0, 1e3), default=1e1
        - laplace_loss_weight: float in (1e1, 1e3), default=1e2
    """

    NAME: ClassVar[str] = "sinet"

    lr = Hyperparameter(float, default=1e-3)
    batch_size = Hyperparameter(int, default=128)
    num_epochs = Hyperparameter(int, default=5_000)
    gradient_clipping_value = Hyperparameter(float, default=1.0)
    mse_loss_weight = Hyperparameter(float, default=1e2)
    eikonal_loss_weight = Hyperparameter(float, default=1e1)
    laplace_loss_weight = Hyperparameter(float, default=1e2)
    _was_early_stopped: ClassVar[bool] = False

    @log_input(log, level=logging.DEBUG)
    def __init__(
        self,
        dataset: BaseClimatrixDataset,
        target_domain: Domain,
        *,
        layers: int = 2,
        hidden_dim: int = 64,
        sorting_group_size: int = 16,
        scale: float = 1.5,
        lr: float = 3e-4,
        batch_size: int = 512,
        num_epochs: int = 5_000,
        num_workers: int = 0,
        device: str = "cuda",
        gradient_clipping_value: float | None = None,
        checkpoint: str | os.PathLike | Path | None = None,
        mse_loss_weight: float = 3e3,
        eikonal_loss_weight: float = 5e1,
        laplace_loss_weight: float = 1e2,
        # use_elevation: bool = False, # NOTE: switched off until we have a proper elevation dataset
        validation: float | BaseClimatrixDataset = 0.2,
        patience: int | None = None,
        overwrite_checkpoint: bool = False,
    ) -> None:
        super().__init__(dataset, target_domain)
        use_elevation = False
        if dataset.domain.is_dynamic:
            log.error("SiNET is not yet supported for dynamic datasets.")
            raise ValueError(
                "SiNET is not yet supported for dynamic datasets."
            )
        if device == "cuda" and not torch.cuda.is_available():
            log.error("CUDA is not available on this machine")
            raise ValueError("CUDA is not available on this machine")
        self.device = torch.device(device)
        self.datasets = self._configure_dataset_generator(
            train_coords=dataset.domain.get_all_spatial_points(),
            train_field=dataset.da.values,
            target_coords=target_domain.get_all_spatial_points(),
            val_portion=validation if isinstance(validation, float) else None,
            val_coordinates=(
                (validation.domain.get_all_spatial_points())
                if isinstance(validation, BaseClimatrixDataset)
                else None
            ),
            val_field=(
                (validation.da.values)
                if isinstance(validation, BaseClimatrixDataset)
                else None
            ),
            use_elevation=use_elevation,
        )
        self.layers = layers
        self.hidden_dim = hidden_dim
        self.sorting_group_size = sorting_group_size
        self.scale = scale
        self.num_workers = num_workers
        self.num_epochs = num_epochs
        self.lr = lr
        self.batch_size = batch_size
        self.gradient_clipping_value = gradient_clipping_value
        self.checkpoint = None
        self.is_model_loaded: bool = False

        self.mse_loss_weight = mse_loss_weight
        self.eikonal_loss_weight = eikonal_loss_weight
        self.laplace_loss_weight = laplace_loss_weight

        self.overwrite_checkpoint = overwrite_checkpoint
        if checkpoint:
            self.checkpoint = Path(checkpoint).expanduser().absolute()
            log.info("Using checkpoint path: %s", self.checkpoint)

        self.patience = patience

    @staticmethod
    def _configure_dataset_generator(
        train_coords: np.ndarray,
        train_field: np.ndarray,
        target_coords: np.ndarray,
        val_portion: float | None = None,
        val_coordinates: np.ndarray | None = None,
        val_field: np.ndarray | None = None,
        use_elevation: bool = False,
    ) -> SiNETDatasetGenerator:
        """
        Configure the SiNET dataset generator.
        """
        log.debug("Configuring SiNET dataset generator...")
        if val_portion is not None and (
            val_coordinates is not None or val_field is not None
        ):
            log.error(
                "Cannot use both `val_portion` and `val_coordinates`/`val_field`."
            )
            raise ValueError(
                "Cannot use both `val_portion` and `val_coordinates`/`val_field`."
            )
        kwargs = {
            "spatial_points": train_coords,
            "field": train_field,
            "target_coordinates": target_coords,
            "degree": True,
            "radius": 1.0,
            "use_elevation": use_elevation,
        }
        if val_portion is not None:
            if not (0 < val_portion < 1):
                log.error("Validation portion must be in the range (0, 1).")
                raise ValueError(
                    "Validation portion must be in the range (0, 1)."
                )
            log.debug("Using validation portion: %0.2f", val_portion)
            kwargs["val_portion"] = val_portion
        elif val_coordinates is not None and val_field is not None:
            log.debug("Using validation coordinates and field for validation.")
            if val_coordinates.shape[0] != val_field.shape[0]:
                log.error(
                    "Validation coordinates and field must have the same number of points."
                )
                raise ValueError(
                    "Validation coordinates and field must have the same number of points."
                )
            kwargs["validation_coordinates"] = val_coordinates
            kwargs["validation_field"] = val_field

        return SiNETDatasetGenerator(**kwargs)

    def _configure_optimizer(
        self, siren_model: torch.nn.Module
    ) -> torch.optim.Optimizer:
        log.info(
            "Configuring Adam optimizer with learning rate: %0.6f",
            self.lr,
        )
        return torch.optim.Adam(lr=self.lr, params=siren_model.parameters())

    def _init_model(self) -> torch.nn.Module:
        log.info("Initializing SiNET model...")
        return SiNET(
            in_features=self.datasets.n_features,
            out_features=1,
            layers=self.layers,
            hidden_dim=self.hidden_dim,
            sorting_group_size=self.sorting_group_size,
            scale=self.scale,
            bias=True,
        ).to(self.device)

    def _maybe_clip_grads(self, siren_model: torch.nn.Module) -> None:
        if self.gradient_clipping_value:
            nn.utils.clip_grad_norm_(
                siren_model.parameters(), self.gradient_clipping_value
            )

    @log_input(log, level=logging.DEBUG)
    def _aggregate_loss(self, loss_component: LossEntity) -> torch.Tensor:
        """
        Aggregate SiNET training loss component.

        Parameters
        ----------
        loss_component : LossEntity
            The losses to be aggregated.

        Returns
        -------
        torch.Tensor
            The aggregated loss.
        """
        return (
            loss_component.mse * self.mse_loss_weight
            + loss_component.eikonal * self.eikonal_loss_weight
            + loss_component.laplace * self.laplace_loss_weight
        )

    def _maybe_load_checkpoint(
        self, siren_model: nn.Module, checkpoint: str | os.PathLike | Path
    ) -> nn.Module:
        if (
            not self.overwrite_checkpoint
            and checkpoint
            and checkpoint.exists()
        ):
            log.debug("Loading checkpoint from %s...", checkpoint)
            try:
                siren_model.load_state_dict(
                    torch.load(checkpoint, map_location=self.device)
                )
                self.is_model_loaded = True
                log.debug("Checkpoint loaded successfully.")
            except RuntimeError as e:
                log.error("Error loading checkpoint: %s.", e)
                raise e
        log.debug(
            "No checkpoint provided or checkpoint not found at %s.", checkpoint
        )
        return siren_model.to(self.device)

    def _maybe_save_checkpoint(
        self, siren_model: nn.Module, checkpoint: Path
    ) -> None:
        if checkpoint:
            if not checkpoint.parent.exists():
                log.debug(
                    "Creating checkpoint directory: %s", checkpoint.parent
                )
                checkpoint.parent.mkdir(parents=True, exist_ok=True)
            log.debug("Saving checkpoint to %s...", checkpoint)
            try:
                torch.save(siren_model.state_dict(), checkpoint)
                log.debug("Checkpoint saved successfully.")
            except Exception as e:
                log.error("Error saving checkpoint: %s", e)
        else:
            log.debug(
                "Checkpoint saving skipped as no checkpoint path is provided."
            )

    @torch.no_grad()
    def _find_surface(self, siren_model, dataset) -> np.ndarray:
        log.debug("Finding surface using the trained INR model")
        data_loader = DataLoader(
            dataset,
            batch_size=50_000,
            shuffle=False,
        )
        all_z = []
        log.info("Creating mini-batches for surface reconstruction...")
        for i, (xy, *_) in enumerate(data_loader):
            log.info("Processing mini-batch %d/%d...", i + 1, len(data_loader))
            xy = xy.to(self.device)
            z = siren_model(xy)
            all_z.append(z.cpu().numpy())
        log.info("Surface finding complete. Concatenating results.")
        return np.concatenate(all_z)

    def _single_epoch_pass(
        self,
        data_loader: DataLoader,
        siren_model: nn.Module,
        optimizer: torch.optim.Optimizer | None = None,
    ) -> float:
        epoch_loss = 0.0
        for xy, true_z in data_loader:
            xy = xy.to(self.device)
            true_z = true_z.to(self.device)
            xy = xy.detach().requires_grad_(True)
            pred_z = siren_model(xy)
            loss_component: LossEntity = compute_sdf_losses(
                xy,
                pred_z * self.datasets.field_transformer.data_range_[0]
                + self.datasets.field_transformer.data_min_[0],
                true_z * self.datasets.field_transformer.data_range_[0]
                + self.datasets.field_transformer.data_min_[0],
            )
            loss = self._aggregate_loss(loss_component=loss_component)
            epoch_loss += loss.item()

            if optimizer is None:
                continue
            optimizer.zero_grad()
            loss.backward()
            self._maybe_clip_grads(siren_model)
            optimizer.step()
        return epoch_loss / len(data_loader)

    @raise_if_not_installed("torch")
    def reconstruct(self) -> BaseClimatrixDataset:
        """Reconstruct the sparse dataset using INR."""
        siren_model = self._init_model()
        siren_model = self._maybe_load_checkpoint(siren_model, self.checkpoint)

        early_stopping = EarlyStopping(
            patience=self.patience,
            delta=0.0,
            checkpoint_path=self.checkpoint,
        )
        if not self.is_model_loaded:
            optimizer = self._configure_optimizer(siren_model)
            log.info("Training SiNET model...")
            train_data_loader = DataLoader(
                self.datasets.train_dataset,
                shuffle=True,
                batch_size=self.batch_size,
                pin_memory=True,
                num_workers=self.num_workers,
            )
            val_data_loader = DataLoader(
                self.datasets.val_dataset,
                shuffle=False,
                batch_size=self.batch_size,
                pin_memory=True,
                num_workers=self.num_workers,
            )

            old_val_loss = np.inf
            for epoch in range(1, self.num_epochs + 1):
                siren_model.train()
                log.debug("Starting epoch %d/%d...", epoch, self.num_epochs)
                train_epoch_loss = self._single_epoch_pass(
                    data_loader=train_data_loader,
                    siren_model=siren_model,
                    optimizer=optimizer,
                )
                siren_model.eval()
                log.debug("Evaluating on validation set...")
                val_epoch_loss = self._single_epoch_pass(
                    data_loader=val_data_loader,
                    siren_model=siren_model,
                    optimizer=None,
                )
                log.debug(
                    "Epoch %d/%d: train loss = %0.4f | val loss = %0.4f",
                    epoch,
                    self.num_epochs,
                    train_epoch_loss,
                    val_epoch_loss,
                )
                if val_epoch_loss < old_val_loss:
                    log.debug(
                        "Validation loss improved from %0.4f to %0.4f",
                        old_val_loss,
                        val_epoch_loss,
                    )
                    self._maybe_save_checkpoint(
                        siren_model=siren_model, checkpoint=self.checkpoint
                    )
                    old_val_loss = val_epoch_loss
                if early_stopping.step(
                    val_metric=val_epoch_loss,
                    model=siren_model,
                ):
                    self._was_early_stopped = True
                    log.debug(
                        "Early stopping triggered at epoch %d/%d",
                        epoch,
                        self.num_epochs,
                    )
                    break
        siren_model.eval()
        log.info("Reconstructing target domain...")
        values = self._find_surface(siren_model, self.datasets.target_dataset)
        unscaled_values = self.datasets.field_transformer.inverse_transform(
            values
        )

        log.debug("Reconstruction completed.")
        return BaseClimatrixDataset(
            self.target_domain.to_xarray(unscaled_values, self.dataset.da.name)
        )



import numpy as np
import xarray as xr

import climatrix as cm

dense_domain = cm.Domain.from_lat_lon(
    slice(-10, 20, 0.1), slice(-40, 30, 0.2), kind="dense"
)
sparse_domain = cm.Domain.from_lat_lon(
    np.linspace(0, 100, 10), np.linspace(0, 100, 10), kind="sparse"
)



"""
This module runs experiment of SiNET method

@author: Jakub Walczak, PhD
"""

import csv
import shutil
from functools import partial
from pathlib import Path
from typing import Any, Callable

import xarray as xr
from bayes_opt import BayesianOptimization
from rich.console import Console

import climatrix as cm

console = Console()

INF_LOSS = -1e4

# Setting up the experiment parameters
NAN_POLICY = "resample"
console.print("[bold green]Using NaN policy: [/bold green]", NAN_POLICY)

SEED = 1
console.print("[bold green]Using seed: [/bold green]", SEED)

DSET_PATH = Path(__file__).parent.parent.parent.joinpath("data")
console.print("[bold green]Using dataset path: [/bold green]", DSET_PATH)

OPTIM_INIT_POINTS: int = 50
console.print(
    "[bold green]Using nbr initial points for optimization: [/bold green]",
    OPTIM_INIT_POINTS,
)

OPTIM_N_ITERS: int = 100
console.print(
    "[bold green]Using iterations for optimization[/bold green]", OPTIM_N_ITERS
)

RESULT_DIR: Path = (
    Path(__file__).parent.parent.parent / "results" / "inr" / "sinet"
)
PLOT_DIR: Path = RESULT_DIR / "plots"
PLOT_DIR.mkdir(parents=True, exist_ok=True)
console.print("[bold green]Plots will be saved to: [/bold green]", PLOT_DIR)

METRICS_PATH: Path = RESULT_DIR / "metrics.csv"
console.print(
    "[bold green]Metrics will be saved to: [/bold green]", METRICS_PATH
)

HYPERPARAMETERS_SUMMARY_PATH: Path = RESULT_DIR / "hparams_summary.csv"
console.print(
    "[bold green]Hyperparameters summary will be saved to: [/bold green]",
    HYPERPARAMETERS_SUMMARY_PATH,
)

BOUNDS = {
    "lr": (1e-5, 1e-2),
    "num_epochs": (50, 500),
    "gradient_clipping_value": (1e-4, 1e4),
    "batch_size": (32, 1024),
    "mse_loss_weight": (1e-5, 1),
    "eikonal_loss_weight": (0, 1e-2),
    "laplace_loss_weight": (0, 1e-2),
    "early_stopping_patience": (10, 200),
}
console.print("[bold green]Hyperparameter bounds: [/bold green]", BOUNDS)

EUROPE_BOUNDS = {"north": 71, "south": 36, "west": -24, "east": 35}
EUROPE_DOMAIN = cm.Domain.from_lat_lon(
    lat=slice(EUROPE_BOUNDS["south"], EUROPE_BOUNDS["north"], 0.1),
    lon=slice(EUROPE_BOUNDS["west"], EUROPE_BOUNDS["east"], 0.1),
    kind="dense",
)


def clear_result_dir():
    console.print(
        "[bold red]Clearing result directory for this experiment...[/bold red]"
    )
    shutil.rmtree(RESULT_DIR, ignore_errors=True)


def create_result_dir():
    PLOT_DIR.mkdir(parents=True, exist_ok=True)


def get_all_dataset_idx() -> list[str]:
    return sorted(
        list({path.stem.split("_")[-1] for path in DSET_PATH.glob("*.nc")})
    )


def update_hparams_csv(hparam_path: Path, hparams: dict[str, Any]):
    fieldnames = [
        "dataset_id",
        "lr",
        "num_epochs",
        "gradient_clipping_value",
        "batch_size",
        "mse_loss_weight",
        "eikonal_loss_weight",
        "laplace_loss_weight",
        "early_stopping_patience",
        "use_elevation",
        "opt_loss",
    ]
    if not hparam_path.exists():
        with open(hparam_path, "w") as f:
            writer = csv.DictWriter(f, fieldnames=fieldnames)
            writer.writeheader()
    with open(hparam_path, "a") as f:
        writer = csv.DictWriter(f, fieldnames=fieldnames)
        writer.writerow(hparams)


def update_metric_csv(metrics_path: Path, metrics: dict[str, Any]):
    fieldnames = ["dataset_id", "RMSE", "MAE", "Max Abs Error", "R^2"]
    if not metrics_path.exists():
        with open(metrics_path, "w") as f:
            writer = csv.DictWriter(f, fieldnames=fieldnames)
            writer.writeheader()
    with open(metrics_path, "a") as f:
        writer = csv.DictWriter(f, fieldnames=fieldnames)
        writer.writerow(metrics)


def is_experiment_done(idx: int) -> bool:
    return (PLOT_DIR / f"{idx}_diffs.png").exists()


def run_single_experiment(
    d: str,
    i: int,
    all_samples: int,
    status: Console.status,
    continuous_update: bool = True,
    reconstruct_dense: bool = True,
):
    cm.seed_all(SEED)
    if is_experiment_done(d):
        console.print(
            f"[bold green]Skipping date {d} as it is already done.[/bold green]"
        )
        return
    status.update(
        f"[magenta]Processing date: {d} ({i + 1}/{all_samples})...",
        spinner="bouncingBall",
    )
    train_dset = xr.open_dataset(
        DSET_PATH / f"ecad_obs_europe_train_{d}.nc"
    ).cm
    val_dset = xr.open_dataset(DSET_PATH / f"ecad_obs_europe_val_{d}.nc").cm
    test_dset = xr.open_dataset(DSET_PATH / f"ecad_obs_europe_test_{d}.nc").cm
    status.update(
        f"[magenta]Optimizing hyper-parameters for date: {d} "
        f"({i + 1}/{all_samples})...",
        spinner="bouncingBall",
    )
    finder = cm.optim.HParamFinder(
        "idw",
        train_dset,
        val_dset,
        metric="mae",
        explore=0.3,
        n_iters=OPTIM_N_ITERS,
        bounds=BOUNDS,
        random_seed=SEED,
    )
    result = finder.optimize()
    console.print("[bold yellow]Optimized parameters:[/bold yellow]")
    console.print(
        "[yellow]Learning rate (lr):[/yellow]", result["best_params"]["lr"]
    )
    console.print(
        "[yellow]Number of epochs:[/yellow]",
        result["best_params"]["num_epochs"],
    )
    console.print(
        "[yellow]Gradient clipping value:[/yellow]",
        result["best_params"]["gradient_clipping_value"],
    )
    console.print(
        "[yellow]Batch size:[/yellow]", result["best_params"]["batch_size"]
    )
    console.print(
        "[yellow]MSE loss weight:[/yellow]",
        result["best_params"]["mse_loss_weight"],
    )
    console.print(
        "[yellow]Eikonal loss weight:[/yellow]",
        result["best_params"]["eikonal_loss_weight"],
    )
    console.print(
        "[yellow]Laplace loss weight:[/yellow]",
        result["best_params"]["laplace_loss_weight"],
    )
    console.print(
        "[yellow]Early stopping patience:[/yellow]",
        result["best_params"]["early_stopping_patience"],
    )
    console.print(
        "[yellow]Use elevation:[/yellow]",
        result["best_params"]["use_elevation"],
    )
    console.print("[yellow]Best loss:[/yellow]", result["best_loss"])
    status.update(
        "[magenta]Reconstructing with optimised parameters...",
        spinner="bouncingBall",
    )
    status.update(
        "[magenta]Concatenating train and validation datasets...",
        spinner="bouncingBall",
    )
    train_val_dset = xr.concat([train_dset.da, val_dset.da], dim="point").cm
    reconstructed_dset = train_val_dset.reconstruct(
        test_dset.domain,
        method="sinet",
        lr=result["best_params"]["lr"],
        num_epochs=result["best_params"]["num_epochs"],
        batch_size=result["best_params"]["batch_size"],
        num_workers=0,
        device="cuda",
        gradient_clipping_value=result["best_params"][
            "gradient_clipping_value"
        ],
        mse_loss_weight=result["best_params"]["mse_loss_weight"],
        eikonal_loss_weight=result["best_params"]["eikonal_loss_weight"],
        laplace_loss_weight=result["best_params"]["laplace_loss_weight"],
        patience=result["best_params"]["early_stopping_patience"],
        use_elevation=result["best_params"]["use_elevation"],
    )
    status.update(
        "[magenta]Saving reconstructed dset to "
        f"{PLOT_DIR}/{d}_reconstructed.png...",
        spinner="bouncingBall",
    )
    reconstructed_dset.plot(show=False).get_figure().savefig(
        PLOT_DIR / f"{d}_reconstructed.png"
    )

    status.update(
        "[magenta]Reconstructing to dense Europe domain...",
        spinner="bouncingBall",
    )
    if reconstruct_dense:
        reconstructed_dense = train_val_dset.reconstruct(
            EUROPE_DOMAIN,
            method="sinet",
            lr=result["best_params"]["lr"],
            num_epochs=result["best_params"]["num_epochs"],
            batch_size=result["best_params"]["batch_size"],
            num_workers=0,
            device="cuda",
            gradient_clipping_value=result["best_params"][
                "gradient_clipping_value"
            ],
            mse_loss_weight=result["best_params"]["mse_loss_weight"],
            eikonal_loss_weight=result["best_params"]["eikonal_loss_weight"],
            laplace_loss_weight=result["best_params"]["laplace_loss_weight"],
            patience=result["best_params"]["early_stopping_patience"],
        )
        status.update(
            "[magenta]Saving reconstructed dense dset to "
            f"{PLOT_DIR}/{d}_reconstructed_dense.png...",
            spinner="bouncingBall",
        )
        reconstructed_dense.plot(show=False).get_figure().savefig(
            PLOT_DIR / f"{d}_reconstructed_dense.png"
        )
    status.update(
        "[magenta]Saving test dset to " f"{PLOT_DIR} / {d}_test.png...",
        spinner="bouncingBall",
    )
    test_dset.plot(show=False).get_figure().savefig(PLOT_DIR / f"{d}_test.png")
    status.update("[magenta]Evaluating...", spinner="bouncingBall")
    cmp = cm.Comparison(reconstructed_dset, test_dset)
    cmp.diff.plot(show=False).get_figure().savefig(PLOT_DIR / f"{d}_diffs.png")
    cmp.plot_signed_diff_hist().get_figure().savefig(
        PLOT_DIR / f"{d}_hist.png"
    )
    metrics = cmp.compute_report()
    metrics["dataset_id"] = d
    hyperparams = {
        "dataset_id": d,
        "lr": result["best_params"]["lr"],
        "num_epochs": result["best_params"]["num_epochs"],
        "gradient_clipping_value": result["best_params"][
            "gradient_clipping_value"
        ],
        "batch_size": result["best_params"]["batch_size"],
        "mse_loss_weight": result["best_params"]["mse_loss_weight"],
        "eikonal_loss_weight": result["best_params"]["eikonal_loss_weight"],
        "laplace_loss_weight": result["best_params"]["laplace_loss_weight"],
        "early_stopping_patience": result["best_params"][
            "early_stopping_patience"
        ],
        "opt_loss": result["best_score"],
    }
    if continuous_update:
        console.print("[bold green]Updating metrics file...[/bold green]")
        update_metric_csv(METRICS_PATH, metrics)

        console.print(
            "[bold green]Updating hyperparameters summary...[/bold green]"
        )
        update_hparams_csv(HYPERPARAMETERS_SUMMARY_PATH, hyperparams)

    return (metrics, hyperparams)


def run_all_experiments_sequentially():
    dset_idx = get_all_dataset_idx()
    with console.status("[magenta]Preparing experiment...") as status:
        for i, d in enumerate(dset_idx):
            if is_experiment_done(d):
                console.print(
                    f"[bold green]Skipping date {d} as it is already done.[/bold green]"
                )
                continue
            run_single_experiment(
                d,
                i,
                len(dset_idx),
                status,
                continuous_update=True,
                reconstruct_dense=True,
            )


if __name__ == "__main__":
    # clear_result_dir()
    create_result_dir()
    run_all_experiments_sequentially()



import unittest.mock as mock

import numpy as np
import pytest
import xarray as xr

from climatrix.comparison import Comparison
from climatrix.dataset.base import BaseClimatrixDataset


@pytest.fixture
def sample_datasets():
    """Create sample datasets for testing comparison functionality."""
    lat = np.array([-45.0, 0.0, 45.0])
    lon = np.array([0.0, 180.0, 360.0])
    
    predicted_data = np.arange(9).reshape(3, 3) + 1.0
    predicted_da = xr.DataArray(
        predicted_data,
        coords=[("lat", lat), ("lon", lon)],
        name="predicted",
    )
    predicted_dataset = BaseClimatrixDataset(predicted_da)
    
    true_data = np.arange(9).reshape(3, 3) + 0.5
    true_da = xr.DataArray(
        true_data,
        coords=[("lat", lat), ("lon", lon)],
        name="true",
    )
    true_dataset = BaseClimatrixDataset(true_da)
    
    return predicted_dataset, true_dataset


@pytest.fixture
def sample_sparse_datasets():
    """Create sample sparse datasets for testing comparison functionality."""
    # Create sparse datasets
    points = np.array([[0.0, 0.0], [1.0, 1.0], [2.0, 2.0]])
    
    predicted_values = np.array([1.0, 2.0, 3.0])
    predicted_da = xr.DataArray(
        predicted_values,
        coords=[("point", np.arange(len(predicted_values)))],
        dims=["point"],
        name="predicted",
    ).assign_coords({
        "latitude": ("point", points[:, 0]),
        "longitude": ("point", points[:, 1]),
    })
    predicted_dataset = BaseClimatrixDataset(predicted_da)
    
    true_values = np.array([1.5, 2.5, 3.5])
    true_da = xr.DataArray(
        true_values,
        coords=[("point", np.arange(len(true_values)))],
        dims=["point"],
        name="true",
    ).assign_coords({
        "latitude": ("point", points[:, 0]),
        "longitude": ("point", points[:, 1]),
    })
    true_dataset = BaseClimatrixDataset(true_da)
    
    return predicted_dataset, true_dataset


class TestComparison:
    """Test class for Comparison functionality."""
    
    def test_plot_diff_accepts_parameters(self, sample_datasets):
        """Test that plot_diff method accepts and passes through plotting parameters."""
        predicted_dataset, true_dataset = sample_datasets
        comparison = Comparison(predicted_dataset, true_dataset)
        
        with mock.patch('climatrix.dataset.base.BaseClimatrixDataset.plot') as mock_plot:
            comparison.plot_diff(
                title="Test Title",
                show=True,
                ax=None,
                figsize=(10, 8),
                cmap="viridis"
            )
            
            mock_plot.assert_called_once_with(
                title="Test Title",
                target=None,
                show=True,
                ax=None,
                figsize=(10, 8),
                cmap="viridis"
            )
    
    def test_plot_diff_with_defaults(self, sample_datasets):
        """Test that plot_diff method works with default parameters."""
        predicted_dataset, true_dataset = sample_datasets
        comparison = Comparison(predicted_dataset, true_dataset)
        
        with mock.patch('climatrix.dataset.base.BaseClimatrixDataset.plot') as mock_plot:
            comparison.plot_diff()
            
            mock_plot.assert_called_once_with(
                title=None,
                target=None,
                show=False,
                ax=None
            )
    
    def test_plot_diff_with_ax_parameter(self, sample_datasets):
        """Test that plot_diff method accepts explicit ax parameter."""
        predicted_dataset, true_dataset = sample_datasets
        comparison = Comparison(predicted_dataset, true_dataset)
        
        with mock.patch('climatrix.dataset.base.BaseClimatrixDataset.plot') as mock_plot:
            mock_ax = mock.Mock()
            comparison.plot_diff(ax=mock_ax)
            
            mock_plot.assert_called_once_with(
                title=None,
                target=None,
                show=False,
                ax=mock_ax
            )

    def test_sparse_domain_comparison_basic(self, sample_sparse_datasets):
        """Test basic sparse domain comparison functionality."""
        predicted_dataset, true_dataset = sample_sparse_datasets
        comparison = Comparison(predicted_dataset, true_dataset)
        
        # Should compute differences successfully
        expected_diff = np.array([-0.5, -0.5, -0.5])
        np.testing.assert_array_almost_equal(comparison.diff.da.values, expected_diff)
        
        # Should be able to compute metrics
        rmse = comparison.compute_rmse()
        assert abs(rmse - 0.5) < 1e-10
        
    def test_sparse_domain_comparison_with_distance_threshold(self, sample_sparse_datasets):
        """Test sparse domain comparison with distance threshold."""
        predicted_dataset, true_dataset = sample_sparse_datasets
        
        # Create offset true dataset
        offset_points = np.array([[0.1, 0.1], [1.1, 1.1], [2.1, 2.1]])
        true_values = np.array([1.5, 2.5, 3.5])
        offset_true_da = xr.DataArray(
            true_values,
            coords=[("point", np.arange(len(true_values)))],
            dims=["point"],
            name="true",
        ).assign_coords({
            "latitude": ("point", offset_points[:, 0]),
            "longitude": ("point", offset_points[:, 1]),
        })
        offset_true_dataset = BaseClimatrixDataset(offset_true_da)
        
        # Test with large threshold - should match all points
        comparison_large = Comparison(predicted_dataset, offset_true_dataset, 
                                    distance_threshold=1.0)
        assert len(comparison_large.diff.da.values) == 3
        
        # Test with small threshold - should match no points
        comparison_small = Comparison(predicted_dataset, offset_true_dataset, 
                                    distance_threshold=0.1)
        assert len(comparison_small.diff.da.values) == 0
        
    def test_mixed_domain_error(self, sample_datasets, sample_sparse_datasets):
        """Test that mixing sparse and dense domains raises an error."""
        dense_dataset, _ = sample_datasets
        sparse_dataset, _ = sample_sparse_datasets
        
        with pytest.raises(ValueError, match="Comparison between sparse and dense domains is not supported"):
            Comparison(dense_dataset, sparse_dataset)
            
        with pytest.raises(ValueError, match="Comparison between sparse and dense domains is not supported"):
            Comparison(sparse_dataset, dense_dataset)
            
    def test_distance_threshold_parameter_in_init(self, sample_sparse_datasets):
        """Test that distance_threshold parameter is stored correctly."""
        predicted_dataset, true_dataset = sample_sparse_datasets
        
        # Test without distance threshold
        comparison1 = Comparison(predicted_dataset, true_dataset)
        assert comparison1.distance_threshold is None
        
        # Test with distance threshold
        comparison2 = Comparison(predicted_dataset, true_dataset, distance_threshold=0.5)
        assert comparison2.distance_threshold == 0.5



import xarray as xr

import climatrix as cm

dset = xr.open_dataset("/storage/tul/projects/climatrix/data/era5-land.nc")
europe = (
    dset.cm.to_signed_longitude()
    .subset(north=71, south=36, west=-24, east=35)
    .itime(0)
)
europe = europe.sample_uniform(number=10000, nan="resample")
europe.plot()
sparse = europe.sample_uniform(number=1000, nan="resample")
recon = sparse.reconstruct(europe.domain, method="idw", k=10)
recon.plot()
comp = cm.Comparison(europe, recon)



import importlib.resources
from unittest.mock import MagicMock, call, patch

import numpy as np
import pytest
import xarray as xr

from climatrix.dataset.base import BaseClimatrixDataset
from climatrix.plot.core import Plot


@pytest.fixture
def sample_dense_dataset():
    lat = np.array([-45.0, 0.0, 45.0])
    lon = np.array([0.0, 180.0, 360.0])
    time = np.array(["2020-01-01", "2020-01-02"], dtype="datetime64")
    data = np.random.rand(2, 3, 3)

    da = xr.DataArray(
        data,
        coords=[("time", time), ("lat", lat), ("lon", lon)],
        name="temperature",
    )
    return BaseClimatrixDataset(da)


@pytest.fixture
def sample_sparse_dataset():
    points = np.array([0, 1, 2])
    lat = np.array([-45.0, 0.0, 45.0])
    lon = np.array([0.0, 180.0, 360.0])
    time = np.array(["2020-01-01", "2020-01-02"], dtype="datetime64")
    data = np.random.rand(2, 3)

    da = xr.DataArray(
        data,
        coords={
            "time": time,
            "point": points,
            "lat": ("point", lat),
            "lon": ("point", lon),
        },
        dims=["time", "point"],
        name="temperature",
    )
    return BaseClimatrixDataset(da)


@pytest.fixture
def sample_dataset_with_vertical():
    lat = np.array([-45.0, 0.0, 45.0])
    lon = np.array([0.0, 180.0, -180.0])
    time = np.array(["2020-01-01", "2020-01-02"], dtype="datetime64")
    level = np.array([850, 500, 200])
    data = np.random.rand(2, 3, 3, 3)

    da = xr.DataArray(
        data,
        coords=[("time", time), ("level", level), ("lat", lat), ("lon", lon)],
        name="temperature",
    )
    return BaseClimatrixDataset(da)


@pytest.fixture
def sample_static_dense_dataset():
    lat = np.array([-45.0, 0.0, 45.0])
    lon = np.array([0.0, 180.0, -180])
    data = np.random.rand(3, 3)

    da = xr.DataArray(
        data,
        coords=[("lat", lat), ("lon", lon)],
        name="temperature",
    )
    return BaseClimatrixDataset(da)


class TestPlotInitialization:

    def test_plot_initialization_with_dataset(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)

        np.testing.assert_allclose(
            plot.dataset.da.values,
            sample_dense_dataset.to_signed_longitude().da.values,
        )
        assert plot.app is not None
        assert plot.app.name == "climatrix.plot.core"

    def test_plot_static_folder_setup(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)

        expected_static = importlib.resources.files(
            "climatrix.resources"
        ).joinpath("static")
        assert plot.app.static_folder == str(expected_static)

    def test_routes_are_registered(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)

        rule_endpoints = [
            rule.endpoint for rule in plot.app.url_map.iter_rules()
        ]

        assert "index" in rule_endpoints
        assert "serve_mylibrary_asset" in rule_endpoints
        assert "get_data" in rule_endpoints
        assert "get_metadata" in rule_endpoints


class TestPlotMetadata:

    def test_get_metadata_dense_with_time_and_vertical(
        self, sample_dataset_with_vertical
    ):
        plot = Plot(sample_dataset_with_vertical)
        metadata = plot.get_metadata()

        assert metadata["has_time"] is True
        assert metadata["has_vertical"] is True
        assert metadata["is_sparse"] is False
        assert metadata["time_count"] == 2
        assert metadata["vertical_count"] == 3
        assert metadata["vertical_name"] == "level"
        assert len(metadata["time_values"]) == 2
        assert len(metadata["vertical_values"]) == 3

        assert all(isinstance(tv, str) for tv in metadata["time_values"])
        assert all(
            isinstance(vv, (int, float)) for vv in metadata["vertical_values"]
        )

    def test_get_metadata_sparse_dataset(self, sample_sparse_dataset):
        plot = Plot(sample_sparse_dataset)
        metadata = plot.get_metadata()

        assert metadata["has_time"] is True
        assert metadata["has_vertical"] is False
        assert metadata["is_sparse"] is True
        assert metadata["time_count"] == 2
        assert "vertical_count" not in metadata
        assert "vertical_name" not in metadata
        assert "vertical_values" not in metadata

    def test_get_metadata_static_dataset(self, sample_static_dense_dataset):
        plot = Plot(sample_static_dense_dataset)
        metadata = plot.get_metadata()

        assert metadata["has_time"] is False
        assert metadata["has_vertical"] is False
        assert metadata["is_sparse"] is False
        assert "time_count" not in metadata
        assert "time_values" not in metadata


class TestDataPreparation:

    def test_prepare_data_calls_correct_method_for_sparse(
        self, sample_sparse_dataset
    ):
        plot = Plot(sample_sparse_dataset)

        with patch.object(plot, "prepare_sparse_data") as mock_sparse:
            mock_sparse.return_value = {"type": "scatter"}
            result = plot.prepare_data(time_idx=1, vertical_idx=0)

            mock_sparse.assert_called_once_with(1, 0)
            assert result["type"] == "scatter"

    def test_prepare_data_calls_correct_method_for_dense(
        self, sample_dense_dataset
    ):
        plot = Plot(sample_dense_dataset)

        with patch.object(plot, "prepare_dense_data") as mock_dense:
            mock_dense.return_value = {"type": "mesh"}
            result = plot.prepare_data(time_idx=1, vertical_idx=0)

            mock_dense.assert_called_once_with(1, 0, 1, None, None, None, None)
            assert result["type"] == "mesh"

    def test_prepare_sparse_data_structure(self, sample_sparse_dataset):
        plot = Plot(sample_sparse_dataset)
        result = plot.prepare_sparse_data(time_idx=0, vertical_idx=0)

        assert result["type"] == "scatter"
        assert "lats" in result
        assert "lons" in result
        assert "values" in result
        assert "min_val" in result
        assert "max_val" in result

        assert isinstance(result["lats"], list)
        assert isinstance(result["lons"], list)
        assert isinstance(result["values"], list)
        assert isinstance(result["min_val"], float)
        assert isinstance(result["max_val"], float)

        assert len(result["lats"]) == len(result["lons"])
        assert len(result["lats"]) == len(result["values"])

    def test_prepare_dense_data_structure(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)
        result = plot.prepare_dense_data(time_idx=0, vertical_idx=0)

        assert result["type"] == "mesh"
        assert "lats" in result
        assert "lons" in result
        assert "values" in result
        assert "min_val" in result
        assert "max_val" in result

        assert isinstance(result["lats"], list)
        assert isinstance(result["lons"], list)
        assert isinstance(result["values"], list)
        assert isinstance(result["min_val"], float)
        assert isinstance(result["max_val"], float)

        assert all(isinstance(row, list) for row in result["values"])

    def test_prepare_data_with_time_indexing(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)

        result_t0 = plot.prepare_dense_data(time_idx=0, vertical_idx=0)
        result_t1 = plot.prepare_dense_data(time_idx=1, vertical_idx=0)

        assert result_t0["type"] == result_t1["type"]
        assert len(result_t0["lats"]) == len(result_t1["lats"])
        assert len(result_t0["lons"]) == len(result_t1["lons"])

        np.testing.assert_raises(
            AssertionError,
            np.testing.assert_array_equal,
            result_t0["values"],
            result_t1["values"],
        )

    def test_prepare_data_with_vertical_indexing(
        self, sample_dataset_with_vertical
    ):
        plot = Plot(sample_dataset_with_vertical)

        result_l0 = plot.prepare_dense_data(time_idx=0, vertical_idx=0)
        result_l1 = plot.prepare_dense_data(time_idx=0, vertical_idx=1)

        assert result_l0["type"] == result_l1["type"]
        assert len(result_l0["lats"]) == len(result_l1["lats"])
        assert len(result_l0["lons"]) == len(result_l1["lons"])

        np.testing.assert_raises(
            AssertionError,
            np.testing.assert_array_equal,
            result_l0["values"],
            result_l1["values"],
        )

    def test_prepare_data_handles_static_dataset(
        self, sample_static_dense_dataset
    ):
        plot = Plot(sample_static_dense_dataset)
        result = plot.prepare_dense_data(time_idx=0, vertical_idx=0)

        assert result["type"] == "mesh"
        assert "lats" in result
        assert "lons" in result
        assert "values" in result
        assert isinstance(result["min_val"], float)
        assert isinstance(result["max_val"], float)

    def test_prepare_data_handles_edge_cases(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)

        result = plot.prepare_dense_data(time_idx=0, vertical_idx=0)
        assert result is not None

    def test_min_max_values_are_computed_correctly(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)
        result = plot.prepare_dense_data(time_idx=0, vertical_idx=0)

        data_slice = sample_dense_dataset.da.values
        expected_min = float(np.min(data_slice))
        expected_max = float(np.max(data_slice))

        assert result["min_val"] == expected_min
        assert result["max_val"] == expected_max


class TestPlotHTMLTemplate:

    @patch("importlib.resources.files")
    def test_get_html_template_reads_correct_file(
        self, mock_files, sample_dense_dataset
    ):
        mock_template_content = "<html>Test Template</html>"
        mock_path = MagicMock()
        mock_path.read_text.return_value = mock_template_content
        mock_files.return_value.joinpath.return_value = mock_path

        plot = Plot(sample_dense_dataset)
        mock_files.assert_called_once_with("climatrix.resources")
        result = plot.get_html_template()

        assert mock_files.call_count == 2
        assert (
            call("static", "plot_template.html")
            in mock_files.return_value.joinpath.call_args_list
        )
        mock_path.read_text.assert_called_once()
        assert result == mock_template_content

    def test_get_html_template_returns_string(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)
        result = plot.get_html_template()

        assert isinstance(result, str)
        assert len(result) > 0


class TestPlotFlaskRoutes:

    def test_index_route_returns_html(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)

        with plot.app.test_client() as client:
            response = client.get("/")
            assert response.status_code == 200
            assert response.content_type.startswith("text/html")

    def test_api_metadata_route_returns_json(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)

        with plot.app.test_client() as client:
            response = client.get("/api/metadata")
            assert response.status_code == 200
            assert response.content_type == "application/json"

            data = response.get_json()
            assert isinstance(data, dict)
            assert "has_time" in data
            assert "has_vertical" in data
            assert "is_sparse" in data

    def test_api_data_route_returns_json(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)

        with plot.app.test_client() as client:
            response = client.get("/api/data")
            assert response.status_code == 200
            assert response.content_type == "application/json"

            data = response.get_json()
            assert isinstance(data, dict)
            assert "type" in data
            assert "lats" in data
            assert "lons" in data
            assert "values" in data

    def test_api_data_route_accepts_parameters(
        self, sample_dataset_with_vertical
    ):
        plot = Plot(sample_dataset_with_vertical)

        with plot.app.test_client() as client:
            response = client.get("/api/data?time_idx=1&vertical_idx=2")
            assert response.status_code == 200

            data = response.get_json()
            assert isinstance(data, dict)
            assert "type" in data
            assert "values" in data

    def test_static_asset_route(self, sample_dense_dataset):
        plot = Plot(sample_dense_dataset)

        with plot.app.test_client() as client:
            response = client.get("/climatrix_assets/logo.svg")
            assert response.status_code in [200, 404]


class TestPlotShowMethod:

    @patch("webbrowser.open")
    @patch("threading.Thread")
    @patch("time.sleep")
    def test_show_starts_server_thread(
        self, mock_sleep, mock_thread, mock_browser, sample_dense_dataset
    ):
        plot = Plot(sample_dense_dataset)

        mock_thread_instance = MagicMock()
        mock_thread.return_value = mock_thread_instance

        mock_sleep.side_effect = [None, KeyboardInterrupt()]

        try:
            plot.show(port=5001, debug=False)
        except KeyboardInterrupt:
            pass

        mock_thread.assert_called_once()
        mock_thread_instance.start.assert_called_once()
        mock_thread_instance.daemon = True

        mock_browser.assert_called_once_with("http://localhost:5001")

    @patch("builtins.print")
    @patch("webbrowser.open")
    @patch("threading.Thread")
    @patch("time.sleep")
    def test_show_prints_status_messages(
        self,
        mock_sleep,
        mock_thread,
        mock_browser,
        mock_print,
        sample_dense_dataset,
    ):
        plot = Plot(sample_dense_dataset)

        mock_thread_instance = MagicMock()
        mock_thread.return_value = mock_thread_instance
        mock_sleep.side_effect = [None, KeyboardInterrupt()]

        try:
            plot.show(port=5001)
        except KeyboardInterrupt:
            pass

        print_calls = [call[0][0] for call in mock_print.call_args_list]
        assert any(
            "server started at http://localhost:5001" in msg
            for msg in print_calls
        )
        assert any("Press Ctrl+C to stop" in msg for msg in print_calls)


class TestPlotErrorHandling:

    def test_plot_handles_invalid_time_index_gracefully(
        self, sample_dense_dataset
    ):
        plot = Plot(sample_dense_dataset)

        try:
            result = plot.prepare_dense_data(time_idx=999, vertical_idx=0)
            assert isinstance(result, dict)
        except (IndexError, KeyError):
            pass

    def test_plot_handles_missing_vertical_dimension(
        self, sample_dense_dataset
    ):
        plot = Plot(sample_dense_dataset)

        result = plot.prepare_dense_data(time_idx=0, vertical_idx=999)
        assert isinstance(result, dict)
        assert result["type"] == "mesh"


class TestPlotIntegrationWithDataset:

    def test_plot_works_with_different_coordinate_names(self):
        latitude = np.array([-45.0, 0.0, 45.0])
        longitude = np.array([0.0, 180.0, 360.0])
        data = np.random.rand(3, 3)

        da = xr.DataArray(
            data,
            coords=[("latitude", latitude), ("longitude", longitude)],
            name="temperature",
        )
        dataset = BaseClimatrixDataset(da)
        plot = Plot(dataset)

        result = plot.prepare_dense_data(time_idx=0, vertical_idx=0)
        assert result["type"] == "mesh"
        assert len(result["lats"]) == 3
        assert len(result["lons"]) == 3

    def test_plot_preserves_data_values(self, sample_static_dense_dataset):
        plot = Plot(sample_static_dense_dataset)
        result = plot.prepare_dense_data(time_idx=0, vertical_idx=0)

        original_values = (
            sample_static_dense_dataset.to_signed_longitude().da.values
        )
        result_values = np.array(result["values"])

        np.testing.assert_array_equal(original_values, result_values)

    def test_plot_handles_single_point_sparse_data(self):
        lat = np.array([45.0])
        lon = np.array([90.0])
        data = np.array([25.5])

        da = xr.DataArray(
            data,
            coords={
                "point": [0],
                "lat": ("point", lat),
                "lon": ("point", lon),
            },
            dims=["point"],
            name="temperature",
        )
        dataset = BaseClimatrixDataset(da)
        plot = Plot(dataset)

        result = plot.prepare_sparse_data(time_idx=0, vertical_idx=0)

        assert result["type"] == "scatter"
        assert len(result["lats"]) == 1
        assert len(result["lons"]) == 1
        assert len(result["values"]) == 1
        assert result["lats"][0] == 45.0
        assert result["lons"][0] == 90.0
        assert result["values"][0] == 25.5



from __future__ import annotations

import logging
import warnings
from abc import abstractmethod
from collections import OrderedDict
from enum import StrEnum
from typing import Any, ClassVar, Literal, Self

import numpy as np
import xarray as xr

from climatrix.dataset.axis import Axis, AxisType, Time
from climatrix.exceptions import MissingAxisError
from climatrix.types import Latitude, Longitude
from climatrix.warnings import (
    DomainMismatchWarning,
    TooLargeSamplePortionWarning,
)

_DEFAULT_LAT_RESOLUTION = 0.1
_DEFAULT_LON_RESOLUTION = 0.1

log = logging.getLogger(__name__)


def validate_input(da: xr.Dataset | xr.DataArray):
    if not isinstance(da, (xr.Dataset, xr.DataArray)):
        raise NotImplementedError(
            "At the moment, dataset can be created only based on "
            "xarray.DataArray or single-variable xarray.Dataset "
            f"objects, but provided {type(da).__name__}"
        )


def ensure_single_var(da: xr.Dataset | xr.DataArray) -> xr.DataArray:
    if isinstance(da, xr.Dataset):
        if len(da.data_vars) > 1:
            raise ValueError(
                "Dataset can be created only based on "
                "xarray.DataArray or xarray.Dataset with single variable "
                "objects, but provided xarray.Dataset with multiple "
                "data_vars."
            )
        elif len(da.data_vars) == 1:
            return da[list(da.data_vars.keys())[0]]
    return da


def match_axes(da: xr.DataArray) -> dict[AxisType, str]:
    axes = OrderedDict()
    for dim in da.dims:
        for axis in Axis.get_all_axes():
            if axis.matches(dim):
                axes[axis.type] = Axis(
                    dim, np.atleast_1d(da[dim].values).astype(axis.dtype), True
                )
    for coord in da.coords.keys():
        for axis in Axis.get_all_axes():
            if axis.type in axes:
                # If the axis is already matched by dim, skip it
                continue
            if axis.matches(coord):
                axes[axis.type] = Axis(
                    coord,
                    np.atleast_1d(da[coord].values).astype(axis.dtype),
                    False,
                )

    return axes


def validate_spatial_axes(axis_mapping: dict[AxisType, Axis]):
    for axis in [AxisType.LATITUDE, AxisType.LONGITUDE]:
        if axis not in axis_mapping:
            raise ValueError(f"Dataset has no {axis.name} axis")


def check_is_dense(axis_mapping: dict[AxisType, Axis]) -> bool:
    return (
        axis_mapping[AxisType.LATITUDE].is_dimension
        and axis_mapping[AxisType.LONGITUDE].is_dimension
    )


def ensure_all_numpy_arrays(coords: dict) -> None:
    return {k: np.array(v, ndmin=1) for k, v in coords.items()}


def filter_out_single_value_coord(coords: dict):
    return {k: v for k, v in coords.items() if len(v) > 1}


class SamplingNaNPolicy(StrEnum):
    IGNORE = "ignore"
    RESAMPLE = "resample"
    RAISE = "raise"

    def __missing__(self, value):
        raise ValueError(f"Unknown NaN policy: {value}")

    @classmethod
    def get(cls, value: str | Self):
        if isinstance(value, cls):
            return value
        try:
            return cls[value.upper()]
        except KeyError:
            raise ValueError(f"Unknown Nan policy: {value}")


class DomainBuilder:
    """
    Builder class for creating domains with various axes.

    Supports a fluent interface for configuring domain axes
    and creating sparse or dense domains.
    """

    def __init__(self):
        self._axes_config = {}

    def _add_axis(self, axis_type: AxisType, **kwargs) -> Self:
        """
        Add an axis to the domain configuration.

        Parameters
        ----------
        axis_type : AxisType
            The type of axis to add.
        **kwargs
            Keyword arguments where the key is the axis name and the value
            is the coordinate data (slice, list, or numpy array).

        Returns
        -------
        DomainBuilder
            The builder instance for method chaining.

        Raises
        ------
        ValueError
            If not exactly one keyword argument is provided.
        """
        if len(kwargs) != 1:
            raise ValueError("Exactly one keyword argument must be provided")
        name, values = next(iter(kwargs.items()))
        self._axes_config[axis_type] = (name, values)
        return self

    def vertical(self, **kwargs) -> Self:
        """
        Add a vertical axis to the domain.

        Parameters
        ----------
        **kwargs
            Keyword arguments where the key is the axis name and the value
            is the coordinate data (slice, list, or numpy array).

        Returns
        -------
        DomainBuilder
            The builder instance for method chaining.

        Examples
        --------
        >>> builder.vertical(depth=slice(10, 100, 1))
        >>> builder.vertical(pressure=[1000, 850, 500, 250])
        """
        return self._add_axis(AxisType.VERTICAL, **kwargs)

    def lat(self, **kwargs) -> Self:
        """
        Add a latitude axis to the domain.

        Parameters
        ----------
        **kwargs
            Keyword arguments where the key is the axis name and the value
            is the coordinate data (slice, list, or numpy array).

        Returns
        -------
        DomainBuilder
            The builder instance for method chaining.

        Examples
        --------
        >>> builder.lat(latitude=[1, 2, 3, 4])
        >>> builder.lat(lat=slice(-90, 90, 1))
        """
        return self._add_axis(AxisType.LATITUDE, **kwargs)

    def lon(self, **kwargs) -> Self:
        """
        Add a longitude axis to the domain.

        Parameters
        ----------
        **kwargs
            Keyword arguments where the key is the axis name and the value
            is the coordinate data (slice, list, or numpy array).

        Returns
        -------
        DomainBuilder
            The builder instance for method chaining.

        Examples
        --------
        >>> builder.lon(longitude=[1, 2, 3, 4])
        >>> builder.lon(lon=slice(-180, 180, 1))
        """
        return self._add_axis(AxisType.LONGITUDE, **kwargs)

    def time(self, **kwargs) -> Self:
        """
        Add a time axis to the domain.

        Parameters
        ----------
        **kwargs
            Keyword arguments where the key is the axis name and the value
            is the coordinate data (slice, list, or numpy array).

        Returns
        -------
        DomainBuilder
            The builder instance for method chaining.

        Examples
        --------
        >>> builder.time(time=['2020-01-01', '2020-01-02'])
        >>> builder.time(valid_time=slice('2020-01-01', '2020-12-31'))
        """
        return self._add_axis(AxisType.TIME, **kwargs)

    def _convert_slice_to_array(self, values):
        """Convert slice objects to numpy arrays."""
        if isinstance(values, slice):
            if values.step is None:
                raise ValueError("Slice step must be specified")
            result = np.arange(
                values.start,
                values.stop,
                values.step,
            )
        else:
            result = np.array(values)

        if len(result) == 0:
            raise ValueError("Resulting array is empty")
        return result

    def _validate_sparse_coordinates(self):
        """Validate that all coordinate arrays have the same length for sparse domains."""
        # Check that latitude and longitude are present
        if AxisType.LATITUDE not in self._axes_config:
            raise ValueError("Latitude axis is required")
        if AxisType.LONGITUDE not in self._axes_config:
            raise ValueError("Longitude axis is required")

        # Get coordinate array lengths for spatial coordinates only
        # Vertical and time can be separate dimensions
        spatial_sizes = {}
        for axis_type, (name, values) in self._axes_config.items():
            if axis_type in [AxisType.LATITUDE, AxisType.LONGITUDE]:
                coord_array = self._convert_slice_to_array(values)
                spatial_sizes[axis_type] = len(coord_array)

        # Check that lat and lon have the same length
        if len(set(spatial_sizes.values())) > 1:
            raise ValueError(
                f"For sparse domain, latitude and longitude arrays must have the same length. "
                f"Got sizes: {spatial_sizes}"
            )

    def sparse(self):
        """
        Create a sparse domain from the configured axes.

        Returns
        -------
        Domain
            A sparse domain instance.

        Raises
        ------
        ValueError
            If coordinate arrays have different lengths or required axes are missing.
        """
        self._validate_sparse_coordinates()

        # Convert configurations to coordinate arrays
        coords = {}
        for axis_type, (name, values) in self._axes_config.items():
            coord_array = self._convert_slice_to_array(values)
            if axis_type in [AxisType.TIME, AxisType.VERTICAL]:
                # Time and vertical are typically dimensions in sparse domains
                coords[name] = coord_array
            else:
                # Spatial coordinates (lat/lon) are indexed by point dimension
                coords[name] = ("point", coord_array)

        # Create xarray Dataset and return Domain
        ds = xr.Dataset(coords=coords)
        return Domain(ds)

    def dense(self):
        """
        Create a dense domain from the configured axes.

        Returns
        -------
        Domain
            A dense domain instance.
        """
        # Convert configurations to coordinate arrays
        coords = {}
        for axis_type, (name, values) in self._axes_config.items():
            coord_array = self._convert_slice_to_array(values)
            coords[name] = coord_array

        # Create xarray Dataset and return Domain
        ds = xr.Dataset(coords=coords)
        return Domain(ds)


class Domain:
    """
    Base class for domain objects.

    Attributes
    ----------
    is_sparse : ClassVar[bool]
        Indicates if the domain is sparse or dense.
    _axes : dict[AxisType, Axis]
        Mapping of `AxisType` to the corresponding `Axis` object.
    """

    __slots__ = ("_axes",)
    is_sparse: ClassVar[bool]
    _axes: dict[AxisType, Axis]

    def __new__(cls, xarray_obj: xr.Dataset | xr.DataArray):
        validate_input(xarray_obj)
        da = ensure_single_var(xarray_obj)
        axis_mapping = match_axes(da)
        validate_spatial_axes(axis_mapping)

        if cls is not Domain:
            return super().__new__(cls)
        if check_is_dense(axis_mapping):
            domain = DenseDomain(da)
        else:
            domain = SparseDomain(da)
        domain._axes = axis_mapping
        return domain

    @property
    def axes(self) -> dict[AxisType, Axis]:
        """
        Get the axes of the domain.

        Returns
        -------
        dict[AxisType, Axis]
            A dictionary mapping `AxisType` to the corresponding `Axis` object.
        """
        return self._axes

    @classmethod
    def from_lat_lon(
        cls,
        lat: slice | np.ndarray = slice(-90, 90, _DEFAULT_LAT_RESOLUTION),
        lon: slice | np.ndarray = slice(-180, 180, _DEFAULT_LON_RESOLUTION),
        kind: Literal["sparse", "dense"] = "dense",
    ) -> Self:
        """
        Create a domain from latitude and longitude coordinates.

        Parameters
        ----------
        lat : slice or np.ndarray
            Latitude coordinates. If a slice is provided, it will be
            converted to a numpy array using the specified step.
        lon : slice or np.ndarray
            Longitude coordinates. If a slice is provided, it will be
            converted to a numpy array using the specified step.
        kind : str
            Type of domain to create. Can be either "dense" or "sparse".
            Default is "dense".

        Returns
        -------
        Domain
            An instance of the Domain class with the specified latitude
            and longitude coordinates.
        """
        if isinstance(lat, slice):
            lat = np.arange(
                lat.start,
                lat.stop + lat.step,
                lat.step,
            )
        if isinstance(lon, slice):
            lon = np.arange(
                lon.start,
                lon.stop + lon.step,
                lon.step,
            )

        # Check if slices produced empty arrays
        if len(lat) == 0:
            raise ValueError(
                "Latitude slice produced an empty array. "
                "Check that the slice parameters (start, stop, step) are valid. "
                "For example, slice(10, 100, -1) would be invalid because "
                "you cannot go from 10 to 100 with a negative step."
            )
        if len(lon) == 0:
            raise ValueError(
                "Longitude slice produced an empty array. "
                "Check that the slice parameters (start, stop, step) are valid. "
                "For example, slice(10, 100, -1) would be invalid because "
                "you cannot go from 10 to 100 with a negative step."
            )
        if kind == "dense":
            return cls(xr.Dataset(coords={"lat": lat, "lon": lon}))
        elif kind == "sparse":
            if len(lat) != len(lon):
                raise ValueError(
                    "For sparse domain, lat and lon must have the same length"
                )
            return cls(
                xr.Dataset(
                    coords={"lat": ("point", lat), "lon": ("point", lon)}
                )
            )
        else:
            raise ValueError(f"Unknown kind: {kind}")

    @classmethod
    def from_axes(cls) -> DomainBuilder:
        """
        Create a domain builder for configuring domains with multiple axes.

        Returns
        -------
        DomainBuilder
            A builder instance for creating domains with various axes.

        Examples
        --------
        >>> domain = (Domain.from_axes()
        ...           .vertical(depth=slice(10, 100, 1))
        ...           .lat(latitude=[1,2,3,4])
        ...           .lon(longitude=[1,2,3,4])
        ...           .sparse())
        >>> domain = (Domain.from_axes()
        ...           .lat(lat=slice(-90, 90, 1))
        ...           .lon(lon=slice(-180, 180, 1))
        ...           .time(time=['2020-01-01', '2020-01-02'])
        ...           .dense())
        """
        return DomainBuilder()

    @property
    def latitude(self) -> Axis:
        """Latitude axis"""
        if AxisType.LATITUDE not in self._axes:
            raise MissingAxisError(
                f"Latitude axis not found in axes ({list(self._axes.keys())})"
            )
        return self._axes[AxisType.LATITUDE]

    @property
    def longitude(self) -> Axis:
        """Longitude axis"""
        if AxisType.LONGITUDE not in self._axes:
            raise MissingAxisError(
                f"Longitude axis not found in axes ({list(self._axes.keys())})"
            )
        return self._axes[AxisType.LONGITUDE]

    @property
    def point(self) -> Axis | None:
        """Point axis"""
        if AxisType.POINT not in self._axes:
            raise MissingAxisError(
                f"Point axis not found in axes ({list(self._axes.keys())})"
            )
        return self._axes.get(AxisType.POINT)

    @property
    def time(self) -> Axis | None:
        """Time axis"""
        if AxisType.TIME not in self._axes:
            raise MissingAxisError(
                f"Time axis not found in axes ({list(self._axes.keys())})"
            )
        return self._axes.get(AxisType.TIME)

    @property
    def vertical(self) -> Axis | None:
        """Vertical axis"""
        if AxisType.VERTICAL not in self._axes:
            raise MissingAxisError(
                f"Vertical axis not found in axes ({list(self._axes.keys())})"
            )
        return self._axes.get(AxisType.VERTICAL)

    @property
    def all_axes_types(self) -> list[AxisType]:
        """All axis types in the domain."""
        return list(self._axes.keys())

    @property
    def dims(self) -> tuple[AxisType, ...]:
        """
        Get the dimensions of the dataset.

        Returns
        -------
        tuple[AxisType, ...]
            A tuple of `AxisType` objects representing the dimensions
            of the dataset.

        Notes
        -----
        The dimensions are determined by the axes that are marked as
        dimensions in the domain. E.g. if underlying dataset has
        shape `(5, 10, 20)`, it means there are 3 dimensional axes.
        """
        return tuple([k for k, v in self._axes.items() if v.is_dimension])

    def get_size(self, axis: AxisType | str) -> int:
        """
        Get the size of the specified axis.

        Parameters
        ----------
        axis : AxisType
            The axis for which to get the size.

        Returns
        -------
        int
            The size of the specified axis.
        """
        axis_type = AxisType.get(axis)
        if axis_type not in self._axes:
            return 1
        return self._axes[axis_type].size

    def has_axis(self, axis: AxisType | str) -> bool:
        """
        Check if the specified axis exists in the domain.

        Parameters
        ----------
        axis : AxisType
            The axis type to check.

        Returns
        -------
        bool
            True if the axis exists, False otherwise.
        """
        return AxisType.get(axis) in self._axes

    def get_axis(self, axis: AxisType | str) -> Axis | None:
        """
        Get the name of the specified axis.

        Parameters
        ----------
        axis : AxisType
            The axis type for which to get the name.

        Returns
        -------
        Axis | None
            The Axis object, or None if not found.
        """
        return self._axes.get(AxisType.get(axis))

    @property
    def size(self) -> int:
        """Domain size."""
        if AxisType.POINT in self._axes:
            lat_lon = self.get_size(AxisType.POINT)
        else:
            lat_lon = self.get_size(AxisType.LATITUDE) * self.get_size(
                AxisType.LONGITUDE
            )
        try:
            if self.time:
                lat_lon *= self.get_size(AxisType.TIME)
        except MissingAxisError:
            pass
        return lat_lon

    @property
    def is_dynamic(self) -> bool:
        """If the domain is dynamic."""
        try:
            return self.time and self.get_size(AxisType.TIME) > 1
        except MissingAxisError:
            return False

    @property
    def shape(self) -> tuple[int, ...]:
        """Domain shape."""
        return tuple(
            [axis.size for axis in self._axes.values() if axis.is_dimension]
        )

    def __eq__(self, value: Any) -> bool:
        if not isinstance(value, Domain):
            return False
        self_only = list(set(self._axes.keys()) - set(value._axes.keys()))
        value_only = list(set(value._axes.keys()) - set(self._axes.keys()))
        if len(self_only) > 0 or len(value_only) > 0:
            log.warning(
                f"Domain mismatch: {self_only} in self, {value_only} in value"
            )
            warnings.warn(
                f"Domain mismatch: {self_only} (self), {value_only} (value). "
                "To compare domains, they need to have the same axis. "
                "Remember axis can be removed by using the `drop` method. ",
                DomainMismatchWarning,
            )
            return False
        for k in self._axes.keys():
            if self._axes[k] != value._axes[k]:
                return False

        return True

    @abstractmethod
    def _compute_subset_indexers(
        north: float | None = None,
        south: float | None = None,
        west: float | None = None,
        east: float | None = None,
    ) -> tuple[dict[str, Any], float, float]:
        raise NotImplementedError

    @abstractmethod
    def _compute_sample_uniform_indexers(
        self, portion: float | None = None, number: int | None = None
    ) -> dict[str, int]:
        raise NotImplementedError

    @abstractmethod
    def _compute_sample_no_nans_indexers(
        self,
        portion: float | None = None,
        number: int | None = None,
    ) -> dict[str, int]:
        raise NotImplementedError

    @abstractmethod
    def _compute_sample_normal_indexers(
        self,
        portion: float | None = None,
        number: int | None = None,
        nan: SamplingNaNPolicy | str = "ignore",
        center_point: tuple[Longitude, Latitude] = None,
        sigma: float = 10.0,
    ) -> dict[str, int]:
        raise NotImplementedError

    @abstractmethod
    def get_all_spatial_points(self) -> np.ndarray:
        raise NotImplementedError

    def _get_sampling_points_nbr(
        self, portion: float | None = None, number: int | None = None
    ) -> int:
        if not (portion or number):
            raise ValueError("Either portion or number must be provided")
        if portion and number:
            raise ValueError(
                "Either portion or number must be provided, but not both"
            )
        if (portion and portion > 1.0) or (number and number > self.size):
            warnings.warn(
                "Requesting more than 100% of the data will result in "
                "duplicates and excessive memory usage",
                TooLargeSamplePortionWarning,
            )
        if portion:
            number = int(self.size * portion)
        return number

    @abstractmethod
    def to_xarray(
        self, values: np.ndarray, name: str | None = None
    ) -> xr.DataArray:
        raise NotImplementedError


class SparseDomain(Domain):
    """
    Sparse domain class.

    Supports operations on sparse spatial domain.
    """

    is_sparse: ClassVar[bool] = True

    def get_all_spatial_points(self) -> np.ndarray:
        """
        Get all spatial points in the domain.

        Returns
        -------
        np.ndarray
            An array of shape (n_points, 2) containing the latitude and
            longitude coordinates of all points in the domain.

        Examples
        --------
        >>> points = domain.get_all_spatial_points()
        >>> points
        array([[ 0. , -0.1],
               [ 0. ,  0. ],
               [ 0. ,  0.1],
               ...
        """
        if self.latitude is None or self.longitude is None:
            raise MissingAxisError(
                "Latitude or Longitude axis is not initialized."
            )
        return np.stack((self.latitude.values, self.longitude.values), axis=1)

    def _compute_subset_indexers(
        self,
        north: float | None = None,
        south: float | None = None,
        west: float | None = None,
        east: float | None = None,
    ) -> tuple[dict[str, Any], float, float]:
        if not (north or south or west or east):
            warnings.warn(
                "Subset parameters not provided. Returning the source dataset"
            )
            return type(self)(self.da)
        if north and south and north < south:
            raise ValueError("North must be greater than south")
        if west and east and west > east:
            raise ValueError("East must be greater than west")
        lat_mask = np.logical_and(
            self.latitude.values >= south, self.latitude.values <= north
        )
        lon_mask = np.logical_and(
            self.longitude.values >= west, self.longitude.values <= east
        )
        point_mask = np.logical_and(lat_mask, lon_mask)
        point_idx = np.where(point_mask)[0]
        idx = {self.point.name: point_idx}

        lon_vals = self.longitude.values[lon_mask]
        return idx, lon_vals.min(), lon_vals.max()

    def _compute_sample_uniform_indexers(
        self, portion: float | None = None, number: int | None = None
    ) -> dict[str, int]:
        indices = np.random.choice(
            self.point.size,
            size=self._get_sampling_points_nbr(portion=portion, number=number),
        )
        return {self.point.name: indices}

    def _compute_sample_normal_indexers(
        self,
        portion: float | None = None,
        number: int | None = None,
        center_point: tuple[Longitude, Latitude] = None,
        sigma: float = 10.0,
    ) -> dict[str, int]:
        n = self._get_sampling_points_nbr(portion=portion, number=number)
        if center_point is None:
            center_point = np.array(
                [
                    np.mean(self.latitude.values),
                    np.mean(self.longitude.values),
                ]
            )
        else:
            center_point = np.array(center_point)

        distances = np.sqrt(
            (self.longitude.values - center_point[0]) ** 2
            + (self.latitude.values - center_point[1]) ** 2
        )
        weights = np.exp(-(distances**2) / (2 * sigma**2))
        weights /= weights.sum()

        indices = np.random.choice(
            self.point.values, size=n, p=weights.flatten()
        )
        return {self.point.name: indices}

    def _compute_sample_no_nans_indexers(
        self,
        da: xr.DataArray,
        portion: float | None = None,
        number: int | None = None,
    ) -> dict[str, int]:
        n = self._get_sampling_points_nbr(portion=portion, number=number)
        notnan_da = da[da.notnull()]
        selected_points_idx = np.random.choice(
            notnan_da[self.point.name].values.size, n
        )
        return {
            self.point.name: selected_points_idx,
        }

    def to_xarray(
        self, values: np.ndarray, name: str | None = None
    ) -> xr.DataArray:
        """
        Convert domain to sparse xarray.DataArray.

        The method applies `values` and (optionally) `name` to
        create a new xarray.DataArray object based on the domain.

        Parameters
        ----------
        values : np.ndarray
            The values to be assigned to the DataArray variable.
        name : str, optional
            The name of the DataArray variable.

        Returns
        -------
        xr.DataArray
            The xarray.DataArray single variable object.

        Raises
        ------
        ValueError
            If the shape of `values` does not match the expected shape.

        Examples
        --------
        >>> domain = Domain.from_lat_lon()
        >>> values = np.random.rand(5, 5)
        >>> da = domain.to_xarray(values, name="example")
        >>> isinstance(da, xr.DataArray)
        True
        >>> da.name
        'example'
        """
        target_shape = tuple([self.get_size(axis) for axis in self.dims])
        values = values.reshape(target_shape)
        coords = {
            self.latitude.name: (
                self.point.name,
                self.latitude.values,
            ),
            self.longitude.name: (
                self.point.name,
                self.longitude.values,
            ),
        }
        dim_names = tuple([self.get_axis(axis).name for axis in self.dims])
        for axis in self._axes.values():
            if axis.name in coords:
                continue
            if isinstance(axis, Time):
                coords[axis.name] = axis.values
            else:
                coords[axis.name] = (self.point.name, axis.values)

        dset = xr.DataArray(
            values,
            coords=coords,
            dims=dim_names,
            name=name,
        )
        return dset


class DenseDomain(Domain):
    """
    Dense domain class.

    Supports operations on dense spatial domain.
    """

    is_sparse: ClassVar[bool] = False

    def get_all_spatial_points(self) -> np.ndarray:
        """
        Get all spatial points in the domain.

        Returns
        -------
        np.ndarray
            An array of shape (n_points, 2) containing the latitude and
            longitude coordinates of all points in the domain.

        Examples
        --------
        >>> points = domain.get_all_spatial_points()
        >>> points
        array([[ 0. , -0.1],
               [ 0. ,  0. ],
               [ 0. ,  0.1],
               ...
        """
        lat_grid, lon_grid = np.meshgrid(
            self.latitude.values, self.longitude.values, indexing="ij"
        )
        lat_grid = lat_grid.flatten()
        lon_grid = lon_grid.flatten()
        return np.stack((lat_grid, lon_grid), axis=1)

    def _compute_subset_indexers(
        self,
        north: float | None = None,
        south: float | None = None,
        west: float | None = None,
        east: float | None = None,
    ) -> tuple[dict[str, Any], float, float]:
        if not (north or south or west or east):
            warnings.warn(
                "Subset parameters not provided. Returning the source dataset"
            )
            return type(self)(self.da)
        if north and south and north < south:
            raise ValueError("North must be greater than south")
        if west and east and west > east:
            raise ValueError("East must be greater than west")

        lats = self.latitude.values
        lons = self.longitude.values
        idx = {
            self.latitude.name: (
                np.s_[south:north]
                if np.all(np.diff(lats) >= 0)
                else np.s_[north:south]
            ),
            self.longitude.name: (
                np.s_[west:east]
                if np.all(np.diff(lons) >= 0)
                else np.s_[east:west]
            ),
        }
        if self.longitude.name in idx:
            start = idx[self.longitude.name].start
            stop = idx[self.longitude.name].stop
        else:
            start = idx[self.latitude.name].start
            stop = idx[self.latitude.name].stop
        return idx, start, stop

    def _compute_sample_uniform_indexers(
        self, portion: float | None = None, number: int | None = None
    ) -> dict[str, int]:
        n = self._get_sampling_points_nbr(portion=portion, number=number)
        selected_lats_idx = np.random.choice(self.latitude.values.size, n)
        selected_lons_idx = np.random.choice(self.longitude.values.size, n)
        return {
            self.latitude.name: xr.DataArray(
                selected_lats_idx, dims=[AxisType.POINT]
            ),
            self.longitude.name: xr.DataArray(
                selected_lons_idx, dims=[AxisType.POINT]
            ),
        }

    def _compute_sample_normal_indexers(
        self,
        portion: float | None = None,
        number: int | None = None,
        center_point: tuple[Longitude, Latitude] = None,
        sigma: float = 10.0,
    ) -> dict[str, int]:
        n = self._get_sampling_points_nbr(portion=portion, number=number)
        if center_point is None:
            center_point = np.array(
                [
                    np.mean(self.latitude.values),
                    np.mean(self.longitude.values),
                ]
            )
        else:
            center_point = np.array(center_point)

        x_grid, y_grid = np.meshgrid(
            self.longitude.values, self.latitude.values
        )
        x_grid_idx, y_grid_idx = np.meshgrid(
            np.arange(self.longitude.values.size),
            np.arange(self.latitude.values.size),
        )
        distances = np.sqrt(
            (x_grid - center_point[0]) ** 2 + (y_grid - center_point[1]) ** 2
        )
        weights = np.exp(-(distances**2) / (2 * sigma**2))
        weights /= weights.sum()

        flat_x = x_grid.flatten()

        indices = np.random.choice(len(flat_x), size=n, p=weights.flatten())
        indices_lats = y_grid_idx.flatten()[indices]
        indices_lons = x_grid_idx.flatten()[indices]

        return {
            self.latitude.name: xr.DataArray(
                indices_lats, dims=[AxisType.POINT]
            ),
            self.longitude.name: xr.DataArray(
                indices_lons, dims=[AxisType.POINT]
            ),
        }

    def _compute_sample_no_nans_indexers(
        self,
        da: xr.DataArray,
        portion: float | None = None,
        number: int | None = None,
    ) -> dict[str, int]:
        n = self._get_sampling_points_nbr(portion=portion, number=number)
        stacked = da.stack(**{AxisType.POINT: da.dims})
        idx = np.arange(len(stacked))[stacked.notnull()]
        selected_idx = np.random.choice(idx, n)
        selected_lat_idx = selected_idx // self.longitude.values.size
        selected_lon_idx = selected_idx % self.longitude.values.size
        return {
            self.latitude.name: xr.DataArray(
                selected_lat_idx, dims=[AxisType.POINT]
            ),
            self.longitude.name: xr.DataArray(
                selected_lon_idx, dims=[AxisType.POINT]
            ),
        }

    def to_xarray(
        self, values: np.ndarray, name: str | None = None
    ) -> xr.DataArray:
        """
        Convert domain to dense xarray.DataArray.

        The method applies `values` and (optionally) `name` to
        create a new xarray.DataArray object based on the domain.

        Parameters
        ----------
        values : np.ndarray
            The values to be assigned to the DataArray variable.
        name : str, optional
            The name of the DataArray variable.

        Returns
        -------
        xr.DataArray
            The xarray.DataArray single variable object.

        Raises
        ------
        ValueError
            If the shape of `values` does not match the expected shape.

        Examples
        --------
        >>> domain = Domain.from_lat_lon()
        >>> values = np.random.rand(5, 5)
        >>> da = domain.to_xarray(values, name="example")
        >>> isinstance(da, xr.DataArray)
        True
        >>> da.name
        'example'
        """
        target_shape = tuple([self.get_size(axis) for axis in self.dims])
        values = values.reshape(target_shape)
        coords = {
            self.latitude.name: self.latitude.values,
            self.longitude.name: self.longitude.values,
        }
        dim_names = tuple([self.get_axis(axis).name for axis in self.dims])
        for axis in self._axes.values():
            if axis.name in coords:
                continue
            coords[axis.name] = axis.values

        return xr.DataArray(
            values,
            coords=coords,
            dims=dim_names,
            name=name,
        )



#!/usr/bin/env python3
import numpy as np
import xarray as xr

try:
    import climatrix as cm

    print("✓ Climatrix imported successfully")
except ImportError:
    print("✗ Could not import climatrix. Make sure it's installed.")
    exit(1)


def create_dummy_dense_dataset():
    """Create a dummy dense climate dataset with time and spatial dimensions."""
    print("\nCreating dummy dense dataset...")

    lats = np.linspace(-90, 90, 50)
    lons = np.linspace(-180, 180, 100)
    time = np.arange(
        np.datetime64("2020-01-01"),
        np.datetime64("2020-12-31"),
        np.timedelta64(1, "D"),
    )

    lat_grid, lon_grid = np.meshgrid(lats, lons, indexing="ij")

    np.random.seed(42)

    base_temp = 15 - 0.7 * np.abs(lat_grid)

    temp_data = np.zeros((len(time), len(lats), len(lons)))
    for i, t in enumerate(time):
        day_of_year = (t - np.datetime64("2020-01-01")) / np.timedelta64(
            1, "D"
        )
        seasonal_factor = 10 * np.cos(2 * np.pi * day_of_year / 365.25)

        seasonal_temp = base_temp + seasonal_factor * np.cos(
            np.radians(lat_grid)
        )

        noise = np.random.normal(0, 2, seasonal_temp.shape)
        temp_data[i] = seasonal_temp + noise

    # Create xarray dataset
    ds = xr.Dataset(
        {"temperature": (["time", "latitude", "longitude"], temp_data)},
        coords={"time": time, "latitude": lats, "longitude": lons},
        attrs={
            "title": "Dummy Global Temperature Dataset",
            "description": "Synthetic daily temperature data for demonstration",
        },
    )

    ds.temperature.attrs = {
        "units": "degrees_C",
        "long_name": "Air Temperature",
        "standard_name": "air_temperature",
    }

    print(f"✓ Dense dataset created: {ds.sizes}")
    return ds.cm


def create_dummy_sparse_dataset():
    """Create a dummy sparse climate dataset with station data."""
    print("\nCreating dummy sparse dataset...")

    np.random.seed(123)
    n_stations = 50

    station_lats = np.random.uniform(-90, 90, n_stations)
    station_lons = np.random.uniform(-180, 180, n_stations)
    station_ids = [f"STATION_{i:03d}" for i in range(n_stations)]

    time = np.arange(
        np.datetime64("2020-01-01"),
        np.datetime64("2020-01-31"),
        np.timedelta64(1, "D"),
    )
    depth = np.arange(0, 20)

    precip_data = np.zeros((len(time), len(depth), n_stations))
    for i, (lat, lon) in enumerate(zip(station_lats, station_lons)):
        base_precip = 5 + 10 * np.exp(-0.02 * lat**2)

        for j in range(len(time)):
            for k in range(len(depth)):
                if np.random.random() < 0.3:  # Rain on 30% of days
                    precip_data[j, k, i] = np.random.exponential(base_precip)
                else:
                    precip_data[j, k, i] = 0

    ds = xr.Dataset(
        {"precipitation": (["time", "depth", "station"], precip_data)},
        coords={
            "time": time,
            "depth": depth,
            "station": station_ids,
            "latitude": ("station", station_lats),
            "longitude": ("station", station_lons),
        },
        attrs={
            "title": "Dummy Station Precipitation Dataset",
            "description": "Synthetic station precipitation data for demonstration",
        },
    )

    ds.precipitation.attrs = {
        "units": "mm/day",
        "long_name": "Daily Precipitation",
        "standard_name": "precipitation_flux",
    }

    print(f"✓ Sparse dataset created: {ds.sizes}")
    return ds.cm


if __name__ == "__main__":
    print("Running interactive plotting demonstration...")
    dset_type = int(
        input("Enter dataset type (1: dense/2: sparse): ").strip().lower()
    )
    if dset_type == 1:
        print("Creating and showing dense dataset plot...")
        dset = create_dummy_dense_dataset()
    elif dset_type == 2:
        print("Creating and showing sparse dataset plot...")
        dset = create_dummy_sparse_dataset()
    else:
        print("Invalid option. Please run again and choose 1 or 2.")
        exit(1)

    plot = cm.plot.Plot(dset)
    plot.show(port=5000)



import xarray as xr

import climatrix as cm

dset = xr.open_dataset("data.nc")
europe = (
    dset.cm.to_signed_longitude()
    .subset(north=71, south=36, west=-24, east=35)
    .itime(0)
)

sparse = europe.sample_uniform(number=10_000, nan="resample")

sparse.plot()

dense = sparse.reconstruct(
    europe.domain,
    method="siren",
)
dense.plot()

without_nans = dense.mask_nan(europe)
without_nans.plot()

comp = cm.Comparison(dense, europe)

comp.save_report("./results/")
